<?php

namespace Drupal\skilling\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\skilling\Access\FilterUserInputInterface;
use Drupal\skilling\Access\SkillingAjaxSecurityInterface;
use Drupal\skilling\SkillingCurrentUser;
use Drupal\skilling\SkillingParser\SkillingParser;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\skilling\SkillingConstants;
use Drupal\skilling\History;
use Drupal\skilling\Utilities as SkillingUtilities;
use Symfony\Component\HttpFoundation\File\Exception\AccessDeniedException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Component\Utility\Html;

/**
 * Class FibController.
 */
class FibController extends ControllerBase {

  /**
   * Entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Skilling current user service.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $skillingCurrentUser;

  /**
   * The Skilling history event recorder.
   *
   * @var \Drupal\skilling\History
   */
  protected $historyService;

  /**
   * The Skilling utilities service.
   *
   * @var \Drupal\Skilling\Utilities
   */
  protected $skillingUtilities;

  /**
   * The Skilling parser, for parsing student solution text.
   *
   * @var \Drupal\skilling\SkillingParser\SkillingParser
   */
  protected $parser;

  /**
   * Service to check AJAX calls.
   *
   * @var \Drupal\skilling\Access\SkillingAjaxSecurityInterface
   */
  protected $ajaxSecurityService;

  /**
   * Service to filter user input.
   *
   * @var \Drupal\skilling\Access\FilterUserInputInterface
   */
  protected $filterInputService;

  /**
   * Constructs a new controller object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity type manager service.
   * @param \Drupal\skilling\SkillingCurrentUser $skilling_current_user
   *   Skilling current user service.
   * @param \Drupal\skilling\History $historyService
   *   Skilling history service.
   * @param \Drupal\skilling\Utilities $skillingUtilities
   *   Skilling utilities service.
   * @param \Drupal\skilling\SkillingParser\SkillingParser $parser
   *   Skilling authoring language parser.
   * @param \Drupal\skilling\Access\SkillingAjaxSecurityInterface $ajaxSecurityService
   *   Ajax security checking service.
   * @param \Drupal\skilling\Access\FilterUserInputInterface $filterInputService
   *   Input filter service.
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    SkillingCurrentUser $skilling_current_user,
    History $historyService,
    SkillingUtilities $skillingUtilities,
    SkillingParser $parser,
    SkillingAjaxSecurityInterface $ajaxSecurityService,
    FilterUserInputInterface $filterInputService
  ) {
    $this->entityTypeManager = $entityTypeManager;
    $this->skillingCurrentUser = $skilling_current_user;
    $this->historyService = $historyService;
    $this->skillingUtilities = $skillingUtilities;
    $this->parser = $parser;
    $this->ajaxSecurityService = $ajaxSecurityService;
    $this->filterInputService = $filterInputService;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /* @noinspection PhpParamsInspection */
    return new static(
      $container->get('entity_type.manager'),
      $container->get('skilling.skilling_current_user'),
      $container->get('skilling.history'),
      $container->get('skilling.utilities'),
      $container->get('skilling.skillingparser'),
      $container->get('skilling.ajax_security'),
      $container->get('skilling.filter_user_input')
    );
  }

  /**
   * Check student response.
   *
   * Expect in request:
   *   internalName: internal name of the item.
   *   response: What the user typed.
   *
   * Return JSON object:
   *  status: "OK" or "error"
   *  errorMessage: message to show if there was a problem.
   *  match: Was there a match?
   *  responseMessage: Message to show about the response.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   HTTP request.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The JSON response to the client.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\skilling\Exception\SkillingParserException
   */
  public function checkResponse(Request $request) {
    $passedBasicSecurity = $this->ajaxSecurityService->securityCheckAjaxRequest(
      $request,
      ['POST'],
      ['/skilling/check-fib-response'],
      TRUE
    );
    if (!$passedBasicSecurity) {
      throw new AccessDeniedException('Access denied, sec fail in FiB::checkResponse');
    }
    // Get the internal name of the MCQ.
    $internalName = $request->get('internalName');
    // Filter with Stripping.
    $internalName = $this->filterInputService->filterUserContent($internalName);
    // Check that FiB with that name exists.
    $fibIds = $this->entityTypeManager->getStorage('node')
      ->getQuery()
      // Is FiB.
      ->condition('type', SkillingConstants::FILL_IN_THE_BLANK_CONTENT_TYPE)
      // Has the right internal name.
      ->condition(SkillingConstants::FIELD_INTERNAL_NAME, $internalName)
      ->execute();
    if (!$fibIds) {
      $message = 'Bad internal name: ' . Html::escape($internalName);
      return $this->formatJsonErrorReport($message);
    }
    // Have the id of the FiB.
    $fibId = reset($fibIds);
    // Get the response the student typed.
    $userResponse = $request->get('answer');
    // Filter.
    $userResponse = $this->filterInputService->filterUserContent($userResponse);
    // Load the FiB.
    /** @var \Drupal\node\NodeInterface $fib */
    $fib = $this->entityTypeManager->getStorage('node')->load($fibId);
    // Get the user who owns the data.
    if (!$fib) {
      $message = 'Bad node in FiB::checkResponse';
      return $this->formatJsonErrorReport($message);
    }
    // Check the user's response.
    /* @noinspection PhpUndefinedFieldInspection */
    $responses = $fib->field_fib_responses;
    $match = FALSE;
    $responseMessage = '';
    foreach ($responses as $response) {
      if ($match) {
        break;
      }
      $responseParagraph = $response->entity;
      $responseType = $responseParagraph->field_response_type->value;
      if ($responseType === 'number') {
        if (is_numeric($userResponse)) {
          $userNumberResponse = (float) $userResponse;
          $minMatchResponse = $responseParagraph->field_lowest_number->value;
          // Was a min specified?
          if (!$minMatchResponse === '0' && !$minMatchResponse) {
            $message = 'Bad min in FiB::checkResponse, fib id:' . Html::escape($fib->id());
            return $this->formatJsonErrorReport($message);
          }
          $minMatchResponse = (float) $minMatchResponse;
          $maxMatchResponse = $responseParagraph->field_highest_number->value;
          if (!$maxMatchResponse === '0' && !$maxMatchResponse) {
            $message = 'Bad max in FiB::checkResponse, fib id:' . Html::escape($fib->id());
            return $this->formatJsonErrorReport($message);
          }
          $maxMatchResponse = (float) $maxMatchResponse;
          if ($userNumberResponse >= $minMatchResponse && $userNumberResponse <= $maxMatchResponse) {
            $match = TRUE;
            $responseMessage = $this->parser->parse($responseParagraph->field_explanation->value);
            break;
          }
        } // End userResponse is_numeric().
      } // End responseType is number.
      elseif ($responseType === 'text') {
        $matchTextItems = $responseParagraph->field_matching_text_responses;
        // Case sensitive applies for each response individually, not for
        // entire question.
        $caseSensitive = $responseParagraph->field_case_sensitive->value;
        $userResponseToCheck = $userResponse;
        if (!$caseSensitive) {
          $userResponseToCheck = strtolower($userResponseToCheck);
        }
        foreach ($matchTextItems as $matchTextItem) {
          $matchText = trim($matchTextItem->value);
          if (!$caseSensitive) {
            $matchText = strtolower($matchText);
          }
          if ($userResponseToCheck == $matchText) {
            $match = TRUE;
            $responseMessage = $this->parser->parse($responseParagraph->field_explanation->value);
            break;
          }
        }
      }
      else {
        $message = 'Bad type in FiB::checkResponse: ' . Html::escape($responseType);
        return $this->formatJsonErrorReport($message);
      }
    } // End for each response.
    if (!$match) {
      // Send no-match response.
      /* @noinspection PhpUndefinedFieldInspection */
      $responseMessage = $this->parser->parse($fib->field_no_match_response->value);
    }
    if ($this->skillingCurrentUser->isStudent()) {
      // Record the response in history. Encode.
//      try {
//        $this->historyService->recordFibAnswer($fibId, $internalName, $userResponse);
//      }
//      catch (\Exception $e) {
//        $message = 'Exception FiB history call, FiB nid: ' . Html::escape($fibId)
//          . $e->getMessage();
//        return $this->formatJsonErrorReport($message);
//      }
    }
    // Send result of response check to user.
    // Response is escaped, maybe should not be, since it comes from authors.
    $result = [
      'status' => 'OK',
      'match' => $match,
      'responseMessage' => $responseMessage,
    ];
    return new JsonResponse($result);
  }

  /**
   * Format an error message for sending back to fib.js.
   *
   * @param string $message
   *   The message.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON result.
   */
  protected function formatJsonErrorReport($message) {
    $message = trim(Html::escape($message));
    $result = new JsonResponse(
      ['status' => 'error', 'errorMessage' => $message]
    );
    return $result;
  }

}
