<?php

namespace Drupal\skilling\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\skilling\Access\FilterUserInputInterface;
use Drupal\skilling\Access\SkillingAjaxSecurityInterface;
use Drupal\skilling\History;
use Drupal\skilling\SkillingClass;
use Drupal\skilling\SkillingConstants;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\skilling\CurrentClass;
use Drupal\skilling\SkillingCurrentUser;
use Drupal\skilling\Utilities;

/**
 * Shows "Your class" page.
 */
class YourClassController extends ControllerBase {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Skilling current class service.
   *
   * @var \Drupal\skilling\CurrentClass
   */
  protected $skillingCurrentClass;

  /**
   * The Skilling current user service.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $skillingCurrentUser;

  /**
   * The Skilling utilities service.
   *
   * @var \Drupal\skilling\Utilities
   */
  protected $skillingUtilities;

  /**
   * The Skilling history event recorder.
   *
   * @var \Drupal\skilling\History
   */
  protected $historyService;

  /**
   * Service to check AJAX calls.
   *
   * @var \Drupal\skilling\Access\SkillingAjaxSecurityInterface
   */
  protected $ajaxSecurityService;

  /**
   * Service to filter user input.
   *
   * @var \Drupal\skilling\Access\FilterUserInputInterface
   */
  protected $filterInputService;

  /**
   * The Skilling class service.
   *
   * @var \Drupal\skilling\SkillingClass
   */
  protected $skillingClass;

  /**
   * Constructs a new MyClassController object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\skilling\CurrentClass $skilling_current_class
   *   The Skilling current class service.
   * @param \Drupal\skilling\SkillingCurrentUser $skilling_current_user
   *   The Skilling current user service.
   * @param \Drupal\skilling\Utilities $skilling_utilities
   *   The Skilling utilities service.
   * @param \Drupal\skilling\History $historyService
   *   The Skilling history event recorder.
   * @param \Drupal\skilling\Access\SkillingAjaxSecurityInterface $ajaxSecurityService
   *   AJAX security checking service.
   * @param \Drupal\skilling\Access\FilterUserInputInterface $filterInputService
   * @param \Drupal\skilling\SkillingClass $skillingClass
   *   The current class service.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    CurrentClass $skilling_current_class,
    SkillingCurrentUser $skilling_current_user,
    Utilities $skilling_utilities,
    History $historyService,
    SkillingAjaxSecurityInterface $ajaxSecurityService,
    FilterUserInputInterface $filterInputService,
    SkillingClass $skillingClass
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->skillingCurrentClass = $skilling_current_class;
    $this->skillingCurrentUser = $skilling_current_user;
    $this->skillingUtilities = $skilling_utilities;
    $this->historyService = $historyService;
    $this->ajaxSecurityService = $ajaxSecurityService;
    $this->filterInputService = $filterInputService;
    $this->skillingClass = $skillingClass;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /* @noinspection PhpParamsInspection */
    return new static(
      $container->get('entity_type.manager'),
      $container->get('skilling.current_class'),
      $container->get('skilling.skilling_current_user'),
      $container->get('skilling.utilities'),
      $container->get('skilling.history'),
      $container->get('skilling.ajax_security'),
      $container->get('skilling.filter_user_input'),
      $container->get('skilling.class')
    );
  }

  /**
   * Generate the "Your class" page.
   *
   * Security note: filter data originating from instructor. Has already
   * been done by Drupal, but... just because.
   *
   * @return string[]
   *   Renderable array.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\skilling\Exception\SkillingNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingUnknownValueException
   * @throws \Drupal\skilling\Exception\SkillingWrongTypeException
   */
  public function yourClass() {
    if ($this->skillingCurrentUser->isAnonymous()) {
      return [
        '#markup' => $this->t('Please log in.'),
      ];
    }
    // Get classes.
    $classes = $this->skillingCurrentClass->getCurrentClasses();
    if (count($classes) === 0) {
      return [
        '#markup' => $this->t('You are not enrolled in any classes.'),
      ];
    }
    $classesRenderable = [];
    /** @var \Drupal\Node\NodeInterface $class */
    foreach ($classes as $class) {
      $classRenderable = [
        'title' => $this->filterInputService->filterUserContent($class->getTitle()),
        'description' => $this->filterInputService->filterUserContent($class->get('body')->value),
      ];
      $classesRenderable[$class->id()] = $classRenderable;
    }
    // Find instructors for classes.
    $instructors = [];
    /** @var \Drupal\node\Entity\Node $class */
    foreach ($classes as $class) {
      $instructorsForClass = $this->skillingClass->getInstructorsForClass($class);
      foreach ($instructorsForClass as $instructor) {
        if (!isset($instructors[$instructor->id()])) {
          $instructors[$instructor->id()] = $instructor;
        }
      }
    }
    // Make renderables for each instructor.
    $instructorsRenderable = [];
    foreach ($instructors as $instructor) {
      $picture = '';
      // Is there a picture?
      if (!$instructor->user_picture->isEmpty()) {
        // Yes. Show it at the original upload size.
        $picture = $instructor->user_picture->entity->getFileUri();
        $picture = file_create_url($picture);
      }
      $instructorRenderable = [
        'first_name' => $this->filterInputService->filterUserContent(
          $instructor->get(SkillingConstants::FIELD_FIRST_NAME)->value
        ),
        'last_name' => $this->filterInputService->filterUserContent(
          $instructor->get(SkillingConstants::FIELD_LAST_NAME)->value
        ),
        'about' => $this->filterInputService->filterUserContent(
          $instructor->get(SkillingConstants::FIELD_ABOUT)->value
        ),
        'picture' => $picture,
        'uid' => $instructor->id(),
      ];
      $instructorsRenderable[$instructor->id()] = $instructorRenderable;
    }
    // Add event to history.
//    $this->historyService->createHistoryNode(
//      $this->t('View your class'), SkillingConstants::HISTORY_EVENT_VIEW_YOUR_CLASS,
//      $this->skillingCurrentUser->id(), []
//    );
    return [
      '#theme' => 'your_class',
      '#classes' => $classesRenderable,
      '#instructors' => $instructorsRenderable,
    ];
  }

}
