<?php

namespace Drupal\skilling;

use Drupal\Component\Utility\Html;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\skilling\Exception\SkillingException;
use Drupal\user\UserDataInterface;
use Drupal\skilling\Utilities as SkillingUtilities;

/**
 * Knows what the user's current class is.
 *
 * A user can be enrolled in more
 * than one class. A user can have different roles in different classes.
 *
 * If the user is enrolled in more than one class, s/he
 * has to choose one to use for things like choosing which
 * calendar to show.
 *
 * The id of the chosen class is stored in a UserData setting.
 * However, that isn't assumed to be correct, since enrollments
 * can change at any time. The constructor checks the
 * current setting value, and resets it if needed.
 *
 * If there is no current class (e.g., anon, no enrollment),
 * the UserData setting is 0.
 *
 * Besides the UserData setting, the constructor sets up
 * several properties that other code can query.
 *
 * $currentClass - node object for the current class. Null if
 *   there is no current class.
 *
 * $currentClasses - node array of classes current user is enrolled in.
 *
 * $currentEnrollments - array of enrollment nodes.
 */
class CurrentClass {

  const CURRENT_CLASS_KEY = 'current_class_id';

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * User data interface.
   *
   * @var \Drupal\user\UserDataInterface
   */
  protected $userData;

  /**
   * The Skilling utilities service.
   *
   * @var \Drupal\skilling\Utilities
   */
  protected $skillingUtilities;

  /**
   * The Skilling utilities service.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $skillingCurrentUser;

  /**
   * The Skilling class service.
   *
   * @var \Drupal\skilling\SkillingClass
   */
  protected $skillingClass;

  /**
   * The current class.
   *
   * @var \Drupal\node\Entity\Node
   */
  protected $currentClass = NULL;

  /**
   * The current user's current enrollment.
   *
   * Used to set the current class.
   *
   * @var \Drupal\node\NodeInterface
   */
  protected $currentEnrollment = NULL;

  /**
   * Current enrollments.
   *
   * @var \Drupal\node\NodeInterface[]
   */
  protected $currentEnrollments = [];

  /**
   * Current classes.
   *
   * @var \Drupal\node\NodeInterface[]
   */
  protected $currentClasses = [];

  /**
   * CurrentClass constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\user\UserDataInterface $userData
   *   The user data service.
   * @param \Drupal\skilling\Utilities $skillingUtilities
   *   The Skilling utilities service.
   * @param \Drupal\skilling\SkillingCurrentUser $skillingCurrentUser
   *   The Skilling current user service.
   * @param \Drupal\skilling\SkillingClass $skillingClass
   *   The Skilling class service.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    UserDataInterface $userData,
    SkillingUtilities $skillingUtilities,
    SkillingCurrentUser $skillingCurrentUser,
    SkillingClass $skillingClass
  ) {
    // Remember services.
    $this->entityTypeManager = $entityTypeManager;
    $this->userData = $userData;
    $this->skillingUtilities = $skillingUtilities;
    $this->skillingCurrentUser = $skillingCurrentUser;
    $this->skillingClass = $skillingClass;
    // Set the current class.
    if (!$this->skillingCurrentUser->isAnonymous()) {
      // User is logged in.
      // Get enrollments.
      $this->loadCurrentUserEnrollments();
      // Are there any?
      if (count($this->currentEnrollments) === 0) {
        // No enrollments for the user.
        $this->currentClass = NULL;
        $this->currentClasses = [];
        $this->currentEnrollments = [];
        $this->currentEnrollment = NULL;
        $this->userData->set(
          SkillingConstants::MODULE_NAME,
          $this->skillingCurrentUser->id(),
          self::CURRENT_CLASS_KEY,
          0
        );
      }
      else {
        // There are enrollments.
        if (count($this->currentEnrollments) === 1) {
          // Just one. Use the class.
          /** @var \Drupal\node\Entity\Node $enrollment */
          $this->currentEnrollment = current($this->currentEnrollments);
          /* @noinspection PhpUndefinedFieldInspection */
          $this->currentClass = $this->currentEnrollment->field_class->entity;
          $this->currentClasses = [$this->currentClass];
          $this->userData->set(
            SkillingConstants::MODULE_NAME,
            $this->skillingCurrentUser->id(),
            self::CURRENT_CLASS_KEY,
            $this->currentClass->id()
          );
        }
        else {
          // There is more than one enrollment.
          // Get the current id setting.
          $currentClassIdSetting = $this->userData->get(
            SkillingConstants::MODULE_NAME,
            $this->skillingCurrentUser->id(),
            self::CURRENT_CLASS_KEY
          );
          // Is there one?
          if (!$currentClassIdSetting) {
            // There are multiple enrollments, but no current class
            // set. Use the class for the first enrollment.
            /** @var \Drupal\node\Entity\Node $currentEnrollment */
            $this->currentEnrollment = current($this->currentEnrollments);
            /* @noinspection PhpUndefinedFieldInspection */
            $this->currentClass = $this->currentEnrollment->field_class->entity;
            $this->currentClasses = [$this->currentClass];
          }
          else {
            // There are multiple enrollments, and there is a class
            // setting.
            // Is the setting valid?
            $settingIsValid = FALSE;
            foreach ($this->currentEnrollments as $currentEnrollment) {
              /** @var \Drupal\node\Entity\Node $class */
              $class = $currentEnrollment->field_class->entity;
              // Add to the current class list while we are here.
              $this->currentClasses[$class->id()] = $class;
              if ($class->id() === $currentClassIdSetting) {
                // Found the current class setting in the
                // enrolled classes. The setting is good.
                $this->currentClass = $class;
                $this->currentEnrollment = $currentEnrollment;
                $settingIsValid = TRUE;
              }
            }
            if (!$settingIsValid) {
              // Did not find the current setting in the current
              // class list. Use the first class instead.
              reset($this->currentEnrollments);
              $this->currentEnrollment = current($this->currentEnrollments);
              $this->currentClass = $this->currentEnrollment->field_class->entity;
            }
          }
          // Current class is set. Remember its id.
          $this->userData->set(
            SkillingConstants::MODULE_NAME,
            $this->skillingCurrentUser->id(),
            self::CURRENT_CLASS_KEY,
            $this->currentClass->id()
          );
        } //End there is more than one enrollment.
      } //End there are enrollments.
    } //End user is logged in.
  }

  /**
   * Is there a current class?
   *
   * @return bool
   *   True if there is a current class.
   */
  public function isCurrentClass() {
    return !is_null($this->currentClass);
  }

  /**
   * Get the current class.
   *
   * If more than one enrollment,
   * the user has to choose one to be current.
   *
   * @return \Drupal\node\NodeInterface
   *   The class.
   */
  public function getCurrentClass() {
    return $this->currentClass;
  }

  /**
   * Get the current classes.
   *
   * @return \Drupal\node\NodeInterface[]
   *   The classes.
   */
  public function getCurrentClasses() {
    return $this->currentClasses;
  }

  /**
   * Get the id of the current class.
   *
   * @return int
   *   The id.
   */
  public function getCurrentClassId() {
    if (!$this->isCurrentClass()) {
      return 0;
    }
    return $this->currentClass->id();
  }

  /**
   * Get the current user's enrollment in the current class.
   *
   * @return \Drupal\node\NodeInterface
   *   Enrollment node.
   */
  public function getCurrentEnrollment() {
    return $this->currentEnrollment;
  }

  /**
   * Get the current user's enrollments.
   *
   * @return \Drupal\node\NodeInterface[]
   *   Enrollment nodes.
   */
  public function getCurrentEnrollments() {
    return $this->currentEnrollments;
  }

  /**
   * Load the current user's enrollments.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function loadCurrentUserEnrollments() {
    if ($this->skillingCurrentUser->isAnonymous()) {
      $this->currentEnrollments = [];
      return;
    }
    // Is user blocked?
    if ($this->skillingCurrentUser->isBlocked()) {
      $this->currentEnrollments = [];
      return;
    }
    // Load the published enrollments.
    $this->currentEnrollments
      = $this->skillingUtilities->getUserEnrollmentsPublished(
          $this->skillingCurrentUser->id()
        );
  }

  /**
   * Set the current class nid in the user data key storage.
   *
   * @param int $nid
   *   The nid.
   *
   * @throws \Drupal\skilling\Exception\SkillingException
   */
  public function setUserDataCurrentClassNid($nid) {
    // Check that the current user is in the class.
    if (!array_key_exists($nid, $this->currentClasses)) {
      throw new SkillingException(
        Html::escape(
          'Current class switch: User '
          . $this->skillingCurrentUser->id()
          . ' not in class ' . $nid
        ), __FILE__, __LINE__
      );
    }
    $this->userData->set(
      SkillingConstants::MODULE_NAME,
      $this->skillingCurrentUser->id(),
      self::CURRENT_CLASS_KEY,
      $nid
    );
    // Invalidate class selector block cache.
    Cache::invalidateTags([SkillingConstants::BLOCK_CURRENT_CLASS_CACHE_TAG]);
  }

  /**
   * Get the current class' cache tags.
   *
   * @return array|string[]
   *   The tags.
   */
  public function getCacheTags() {
    return $this->currentClass->getCacheTags();
  }

  /**
   * Get the current user's instructor.
   *
   * Return NULL if not a student, no class, or no instructor.
   * If multiple instructors, return one.
   *
   * @return null|\Drupal\user\Entity\User
   *   The instructor.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingWrongTypeException
   */
  public function getCurrentInstructor() {
    if (!$this->skillingCurrentUser->isStudent()) {
      return NULL;
    }
    if (is_null($this->currentClass)) {
      return NULL;
    }
    /** @var \Drupal\user\Entity\User[] $instructors */
    $instructors = $this->skillingClass->getInstructorsForClass($this->currentClass);
    if (count($instructors) === 0) {
      return NULL;
    }
    // If there is more than one, just get the first one.
    /** @var \Drupal\user\Entity\User $instructor */
    $instructor = reset($instructors);
    return $instructor;
  }

}
