<?php

namespace Drupal\skilling\Plugin\HistoryItem;

use Drupal\node\NodeInterface;

/**
 * Represents history event for sending a message from someone to someone else.
 *
 * It's used to show deets for both instructor-to-student and
 * student-to-instructor messages.
 */
class MessageFromSomeoneToSomeone {

  /**
   * Make a render array for deets.
   *
   * @param \Drupal\node\NodeInterface $historyNode
   *   History event node.
   * @param string $templateName
   *   Name of the Twig template to use.
   *
   * @return array
   *   Render array.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\skilling\Exception\SkillingException
   */
  public function getDetailsRenderable(NodeInterface $historyNode, $templateName) {
    list($instructorId, $studentId, $subject, $message) = $this->extractDeets($historyNode);
    $result = [
      '#theme' => $templateName,
      '#subject' => $subject,
      '#message' => $message,
    ];
    /** @var \Drupal\skilling\SkillingCurrentUser $currentUser */
    $currentUser = \Drupal::service('skilling.skilling_current_user');
    if ($currentUser->isAdministrator()) {
      // Admins get more stuff in their deets.
      /** @var \Drupal\skilling\SkillingUserFactory $userFactory */
      $userFactory = \Drupal::service('skilling.skilling_user_factory');
      $student = $userFactory->makeSkillingUser($studentId);
      $instructor = $userFactory->makeSkillingUser($instructorId);
      $result['#instructor_id'] = $instructorId;
      $result['#student_id'] = $studentId;
      $result['#student_name'] = $student->getAccountName();
      $result['#instructor_name'] = $instructor->getAccountName();
    }
    return $result;
  }

  /**
   * Make XML representation of deets.
   *
   * @param \Drupal\node\NodeInterface $historyNode
   *   History event node.
   * @param string $outerTag
   *   Outer tag.
   *
   * @return string
   *   XML representation.
   */
  public function getDetailsXml(NodeInterface $historyNode, $outerTag) {
    list($instructorId, $studentId, $subject, $message)
      = $this->extractDeets($historyNode);
    $instructorId = (int) $instructorId;
    $studentId = (int) $studentId;
    $subject = htmlspecialchars($subject, ENT_XML1 | ENT_QUOTES, 'UTF-8');
    $message = htmlspecialchars($message, ENT_XML1 | ENT_QUOTES, 'UTF-8');
    $result =
       '<' . $outerTag . ' student-id="' . $studentId . '" instructor-id="' . $instructorId . '">'
        . "<subject>$subject</subject>"
        . "<message>$message</message>"
      . "</$outerTag>";
    return $result;
  }

  /**
   * Extract the details into separate variables.
   *
   * @param \Drupal\node\NodeInterface $historyNode
   *   The history node.
   *
   * @return array
   *   Deets in variables.
   */
  protected function extractDeets(NodeInterface $historyNode) {
    $deets = $historyNode->get('field_details')->value;
    // Unserialize it.
    $deets = unserialize($deets);
    $instructorId = $deets['instructor_id'];
    $studentId = $deets['student_id'];
    $subject = $deets['subject'];
    $message = $deets['message'];
    return [$instructorId, $studentId, $subject, $message];
  }

  /**
   * Is the event to be logged in history?
   *
   * Always log.
   *
   * @return bool
   *   Whether to log the event.
   */
  public function isLogged() {
    return TRUE;
  }

}
