<?php

namespace Drupal\skilling;

use DateTime;
use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Utility\Token;
use \Drupal\Core\Render\BubbleableMetadata;
use Drupal\skilling\Exception\SkillingException;

/**
 * Class SkillingTokens.
 */
class SkillingTokens implements SkillingTokensInterface {

  const TRUE_VALUE = 'TRUE';
  const FALSE_VALUE = 'FALSE';
  const NOT_AVAILABLE_VALUE = 'FALSE';

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;
  /**
   * Drupal\skilling\SkillingCurrentUser definition.
   *
   * @var \Drupal\skilling\SkillingCurrentUser
   */
  protected $skillingCurrentUser;
  /**
   * Drupal\skilling\CurrentClass definition.
   *
   * @var \Drupal\skilling\CurrentClass
   */
  protected $currentClass;
  /**
   * Drupal\skilling\Submissions definition.
   *
   * @var \Drupal\skilling\Submissions
   */
  protected $submissionsService;
  /**
   * Drupal\skilling\Utilities definition.
   *
   * @var \Drupal\skilling\Utilities
   */
  protected $skillingUtilities;

  /**
   * The token service.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected $tokenService;

  /**
   * Constructs a new SkillingTokens object.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, SkillingCurrentUser $skillingCurrentUser, CurrentClass $currentClass, Submissions $skilling_submissions, Utilities $skilling_utilities, Token $token) {
    $this->entityTypeManager = $entity_type_manager;
    $this->skillingCurrentUser = $skillingCurrentUser;
    $this->currentClass = $currentClass;
    $this->submissionsService = $skilling_submissions;
    $this->skillingUtilities = $skilling_utilities;
    $this->tokenService = $token;
  }

  /**
   * Define Skilling tokens.
   *
   * @return array
   *   Token definitions.
   */
  public function tokenInfo() {
    $info = [];
    // General Skilling token category.
    // Not specific to exercises, or submissions.
    $info['types']['skilling'] = ['name' => t('Skilling'), 'description' => t('Skilling tokens of doom')];
    // To help authors test tokens.
    $info['tokens']['skilling']['doom-song'] =
      [
        'name' => t('Doom song'),
        'description' => t('A token to sing the Doom Song. So authors can try out the token system.'),
      ];
    // Tokens about exercises.
    $info['types']['exercise'] = ['name' => t('Exercise'), 'description' => t('Exercise tokens.')];
    $info['tokens']['exercise']['submitted'] =
      [
        'name' => t('Submitted'),
        'description' => t('Has the current user submitted a given exercise?'),
        'dynamic' => TRUE,
      ];
    $info['tokens']['exercise']['completed'] =
      [
        'name' => t('Completed'),
        'description' => t('Has the current user completed a given exercise?'),
        'dynamic' => TRUE,
      ];
    // Tokens about classes and calendar events.
    $info['types']['class'] = ['name' => t('Class'), 'description' => t('Class tokens.')];
    $info['tokens']['class']['event-passed'] =
      [
        'name' => t('Event passed'),
        'description' => t('The date for an event is in the past.'),
        'dynamic' => TRUE,
      ];
    return $info;
  }

  /**
   * Alter existing token definitions.
   *
   * @param array $data
   *   Existing token data.
   */
  public function tokenInfoAlter(array &$data) {
    // Add a token to current-user for days since last login.
    $data['tokens']['current-user']['days-since-login'] = [
      'name' => t("Days since current user's last login."),
      'description' => t("Number of days since current user's last login. -1 for never logged in before."),
    ];
    // Add tokens to test for roles.
    $data['tokens']['current-user']['is-author'] = [
      'name' => t('Author role'),
      'description' => t('Does the current user have the author role?'),
    ];
    $data['tokens']['current-user']['is-reviewer'] = [
      'name' => t('Reviewer role'),
      'description' => t('Does the current user have the reviewer role?'),
    ];
    $data['tokens']['current-user']['is-instructor'] = [
      'name' => t('Instructor role'),
      'description' => t('Does the current user have the instructor role?'),
    ];
    $data['tokens']['current-user']['is-grader'] = [
      'name' => t('Grader role'),
      'description' => t('Does the current user have the grader role?'),
    ];
    $data['tokens']['current-user']['is-student'] = [
      'name' => t('Student role'),
      'description' => t('Does the current user have the student role?'),
    ];
    $data['tokens']['current-user']['is-anonymous'] = [
      'name' => t('Anonymous'),
      'description' => t('Is the current user not logged in?'),
    ];
    $data['tokens']['current-user']['number-exercises-with-submissions'] = [
      'name' => t('Number of exercises with submissions'),
      'description' => t('Number of exercises the current user has submitted solutions for. Does not include resubmissions.'),
    ];
    $data['tokens']['current-user']['number-exercises-completed'] = [
      'name' => t('Number of exercises completed'),
      'description' => t('Number of exercises the current user has completed.'),
    ];
    $data['tokens']['current-user']['instructor-first-name'] = [
      'name' => t("Instructor's first name."),
      'description' => t("The first name of the current user's instructor."),
    ];
    $data['tokens']['current-user']['instructor-last-name'] = [
      'name' => t("Instructor's last name."),
      'description' => t("The last name of the current user's instructor."),
    ];
    $data['tokens']['current-user']['instructor-email'] = [
      'name' => t("Instructor's email address."),
      'description' => t("The email address of the current user's instructor."),
    ];
  }

  /**
   * Supply replacement data for Skilling tokens.
   *
   * @param string $type
   *   Token type.
   * @param array $tokens
   *   Array of tokens to be replaced.
   * @param array $data
   *   Associative array of data objects to be used when
   *   generating replacement values.
   * @param array $options
   *   Associative array of options for token replacement.
   * @param \Drupal\Core\Render\BubbleableMetadata $bubbleableMetadata
   *   The bubbleable metadata.
   *
   * @return array
   *   Associative array of replacement values.
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\skilling\Exception\SkillingValueMissingException
   * @throws \Drupal\skilling\Exception\SkillingWrongTypeException
   */
  public function tokens($type, array $tokens, array $data, array $options, BubbleableMetadata $bubbleableMetadata) {
    $replacements = [];
    switch ($type) {
      case 'skilling':
        $replacements = $this->generalTokens($tokens, $data, $options, $bubbleableMetadata);
        break;

      case 'current-user':
        $replacements = $this->currentUserTokens($tokens, $data, $options, $bubbleableMetadata);
        break;

      case 'exercise':
        $replacements = $this->exerciseTokens($tokens, $data, $options, $bubbleableMetadata);
        break;

      case 'class':
        $replacements = $this->classTokens($tokens, $data, $options, $bubbleableMetadata);
        break;

    }
    return $replacements;
  }

  /**
   * Process general tokens.
   *
   * @param array $tokens
   *   Array of tokens to be replaced.
   * @param array $data
   *   Associative array of data objects to be used when
   *   generating replacement values.
   * @param array $options
   *   Associative array of options for token replacement.
   * @param \Drupal\Core\Render\BubbleableMetadata $bubbleableMetadata
   *   The bubbleable metadata.
   * @return array
   *   Associative array of replacement values.
   */
  protected function generalTokens($tokens, array $data, array $options, BubbleableMetadata $bubbleableMetadata) {
    foreach ($tokens as $name => $original) {
      // What's the token?
      switch ($name) {
        case 'doom-song':
          $replacements[$original] = 'Doom de doom doom doom de doooooom dooom.';
          break;
      }
    }
    return $replacements;
  }

  /**
   * Process current user tokens.
   *
   * @param array $tokens
   *   Array of tokens to be replaced.
   * @param array $data
   *   Associative array of data objects to be used when
   *   generating replacement values.
   * @param array $options
   *   Associative array of options for token replacement.
   * @param \Drupal\Core\Render\BubbleableMetadata $bubbleableMetadata
   *   The bubbleable metadata.
   *
   * @return array
   *   Associative array of replacement values.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\skilling\Exception\SkillingValueMissingException
   * @throws \Drupal\skilling\Exception\SkillingWrongTypeException
   */
  protected function currentUserTokens(array $tokens, array $data, array $options, BubbleableMetadata $bubbleableMetadata) {
    $replacements = NULL;
    foreach ($tokens as $name => $original) {
      $name = strtolower($name);
      switch ($name) {
        case 'days-since-login':
          $lastLogin = $this->skillingCurrentUser->getDrupalUser()->getLastLoginTime();
          if (!$lastLogin) {
            $replacements[$original] = 'never';
          }
          else {
            $seconds = time() - $lastLogin;
            $minutes = (int) ($seconds / 60);
            $hours = (int) ($minutes / 60);
            $days = (int) ($hours / 24);
            $replacements[$original] = $days;
          }
          break;

        case 'is-author':
          $replacements[$original] = $this->skillingCurrentUser->isAuthor()
            ? self::TRUE_VALUE : self::FALSE_VALUE;
          break;

        case 'is-reviewer':
          $replacements[$original] = $this->skillingCurrentUser->isReviewer()
            ? self::TRUE_VALUE : self::FALSE_VALUE;
          break;

        case 'is-instructor':
          $replacements[$original] = $this->skillingCurrentUser->isInstructor()
            ? self::TRUE_VALUE : self::FALSE_VALUE;
          break;

        case 'is-grader':
          $replacements[$original] = $this->skillingCurrentUser->isGrader()
            ? self::TRUE_VALUE : self::FALSE_VALUE;
          break;

        case 'is-student':
          $replacements[$original] = $this->skillingCurrentUser->isStudent()
            ? self::TRUE_VALUE : self::FALSE_VALUE;
          break;

        case 'is-anonymous':
          $replacements[$original] = $this->skillingCurrentUser->isAnonymous()
            ? self::TRUE_VALUE : self::FALSE_VALUE;
          break;

        case 'number-exercises-with-submissions':
          if (!$this->skillingCurrentUser->isStudent()) {
            $replacements[$original] = self::NOT_AVAILABLE_VALUE;
          }
          else {
            // Current user is a student.
            $submissionStatuses = $this->submissionsService->getSubmissionStatusesForStudent(
              $this->skillingCurrentUser
            );
            $numberSubmissions = count($submissionStatuses);
            $replacements[$original] = $numberSubmissions;
          }
          break;

        case 'number-exercises-completed':
          if (!$this->skillingCurrentUser->isStudent()) {
            $replacements[$original] = self::NOT_AVAILABLE_VALUE;
          }
          else {
            // Current user is a student.
            $submissionStatuses = $this->submissionsService->getSubmissionStatusesForStudent(
              $this->skillingCurrentUser
            );
            $numberComplete = 0;
            foreach ($submissionStatuses as $submissionStatus) {
              if ($submissionStatus === SkillingConstants::EXERCISE_SUBMISSION_COMPLETE) {
                $numberComplete++;
              }
            }
            $replacements[$original] = $numberComplete;
          }
          break;

        case 'instructor-first-name':
          $replacements[$original] = $this->getInstructorAttribute('first name');
          break;

        case 'instructor-last-name':
          $replacements[$original] = $this->getInstructorAttribute('last name');
          break;

        case 'instructor-email':
          $replacements[$original] = $this->getInstructorAttribute('email');
          break;

      }
    }
    if (!is_null($replacements)) {
      return $replacements;
    }
  }

  /**
   * Get an attribute of the current user's instructor.
   *
   * Return a not-available value when appropriate.
   *
   * @param string $attributeName
   *   Attribute name, e.g., 'first name'.
   *
   * @return string
   *   Attribute value.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingWrongTypeException
   */
  protected function getInstructorAttribute($attributeName) {
    $result = self::NOT_AVAILABLE_VALUE;
    $instructor = $this->currentClass->getCurrentInstructor();
    if ($instructor) {
      switch ($attributeName) {
        case 'first name':
          $result = $instructor->field_first_name->value;
          break;

        case 'last name':
          $result = $instructor->field_last_name->value;
          break;

        case 'email':
          $result = $instructor->getEmail();
          break;
      }
      if (!$result) {
        $result = self::NOT_AVAILABLE_VALUE;
      }
    }
    return $result;
  }

  /**
   * Process exercise tokens.
   *
   * @param array $tokens
   *   Array of tokens to be replaced.
   * @param array $data
   *   Associative array of data objects to be used when
   *   generating replacement values.
   * @param array $options
   *   Associative array of options for token replacement.
   * @param \Drupal\Core\Render\BubbleableMetadata $bubbleableMetadata
   *   The bubbleable metadata.
   *
   * @return array
   *   Associative array of replacement values.
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\skilling\Exception\SkillingNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingUnknownValueException
   * @throws \Drupal\skilling\Exception\SkillingValueMissingException
   * @throws \Drupal\skilling\Exception\SkillingWrongTypeException
   */
  protected function exerciseTokens(array $tokens, array $data, array $options, BubbleableMetadata $bubbleableMetadata) {
    $replacements = NULL;
    foreach ($tokens as $name => $original) {
      $submittedTokens = $this->tokenService->findWithPrefix($tokens, 'submitted');
      if ($submittedTokens) {
        // Submitted token applies to students. For others, return NA value.
        if (!$this->skillingCurrentUser->isStudent()) {
          $replacements[$original] = self::NOT_AVAILABLE_VALUE;
        }
        else {
          // Is a student.
          // Has an exercise been submitted?
          $exerciseInternalName = array_keys($submittedTokens)[0];
          // Find the exercise.
          $exercise = $this->skillingUtilities->getPublishedExerciseWithInternalName($exerciseInternalName);
          if ($exercise) {
            $exerciseNid = $exercise->id();
            // Get the submission status for the exercise.
            $status = $this->submissionsService->getSubmissionStatusForStudentExercise(
              $this->skillingCurrentUser, $exerciseNid
            );
            // Any status but no-submission shows there has been a submission.
            $replacements[$original] = ($status === SkillingConstants::EXERCISE_SUBMISSION_NO_SUBMISSION)
              ? self::FALSE_VALUE : self::TRUE_VALUE;
          }
        }
      }
      $completedTokens = $this->tokenService->findWithPrefix($tokens, 'completed');
      if ($completedTokens) {
        // Completed token applies to students. For others, return NA value.
        if (!$this->skillingCurrentUser->isStudent()) {
          $replacements[$original] = self::NOT_AVAILABLE_VALUE;
        }
        else {
          // Is a student.
          // Has an exercise been completed?
          $exerciseInternalName = array_keys($completedTokens)[0];
          // Find the exercise.
          $exercise = $this->skillingUtilities->getPublishedExerciseWithInternalName($exerciseInternalName);
          if ($exercise) {
            $exerciseNid = $exercise->id();
            // Get the submission status for the exercise.
            $status = $this->submissionsService->getSubmissionStatusForStudentExercise(
              $this->skillingCurrentUser, $exerciseNid
            );
            $replacements[$original] = ($status === SkillingConstants::EXERCISE_SUBMISSION_COMPLETE)
              ? self:: TRUE_VALUE : self::FALSE_VALUE;
          }
        }
      }
    }
    if (!is_null($replacements)) {
      return $replacements;
    }
  }

  /**
   * Process class tokens.
   *
   * @param array $tokens
   *   Array of tokens to be replaced.
   * @param array $data
   *   Associative array of data objects to be used when
   *   generating replacement values.
   * @param array $options
   *   Associative array of options for token replacement.
   * @param \Drupal\Core\Render\BubbleableMetadata $bubbleableMetadata
   *   The bubbleable metadata.
   *
   * @return array
   *   Associative array of replacement values.
   *
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function classTokens(array $tokens, array $data, array $options, BubbleableMetadata $bubbleableMetadata) {
    return;
    $replacements = NULL;
    foreach ($tokens as $name => $original) {
      $eventTokens = $this->tokenService->findWithPrefix($tokens, 'event-passed');
      if ($eventTokens) {
        // Has an event passed?
        $eventInternalName = array_keys($eventTokens)[0];
        // Find an event with that internal name.
        $eventIds = $this->entityTypeManager->getStorage('node')->getQuery()
          ->condition('type', SkillingConstants::EVENT_CONTENT_TYPE)
          ->condition('field_internal_name', $eventInternalName)
          ->condition('status', 1)
          ->execute();
        // Found more than one?
        if (count($eventIds) > 1) {
          throw new SkillingException(
            'Too many events with internal name: ' . $eventInternalName,
            __FILE__, __LINE__
          );
        }
        // Find any?
        if (count($eventIds) === 1) {
          $eventId = reset($eventIds);
          // Find the current class.
          $class = $this->currentClass->getCurrentClass();
          if ($class) {
            // Find the calendar for the current class.
            $calendarId = $class->get('field_calendar')->target_id;
            // Is the event on the calendar?
            $calendarEventIds = $this->entityTypeManager
              ->getStorage('node')->getQuery()
// REMOVING CALENDAR EVENTS.
//              ->condition('type', SkillingConstants::CALENDAR_EVENT_CONTENT_TYPE)
              ->condition('field_calendar.target_id', $calendarId)
              ->condition('field_event.target_id', $eventId)
              ->condition('status', 1)
              ->execute();
            if (count($calendarEventIds) === 1) {
              // Found the event on the current class's calendar.
              $calendarEventId = reset($calendarEventIds);
              $calendarEvent = $this->entityTypeManager
                ->getStorage('node')->load($calendarEventId);
              // How many days from the start of the class is
              // the event scheduled?
              $eventDayOffset = $calendarEvent->field_day->value;
              // Check just in case.
              if ($eventDayOffset) {
                // Get the start of the class.
                $whenCalendarStarts = $class->field_when_starts->value;
                $startDate = new DateTime($whenCalendarStarts);
                // What is today, expressed as number of days from
                // the start of course?
                $now = new DateTime();
                $todayAsDayNumber = $now->diff($startDate)->days + 1;
                // Has the event passed?
                if ($todayAsDayNumber > $eventDayOffset) {
                  $replacements[$original] = self::TRUE_VALUE;
                }
                else {
                  $replacements[$original] = self::FALSE_VALUE;
                }
              } // End there is a day in the calendar event record.
            } // End the event is on the calendar.
          } // End there is a current class.
        } // End an event with the given internal name exists.
      } // End there was an event-passed token.
    } // End for each token.
    if (!is_null($replacements)) {
      return $replacements;
    }
  }

}
